from __future__ import annotations

from typing import TYPE_CHECKING

from . import app_info, base, client_server, client_server_2, friends, leaderboards, login, ucm, user_stats
from .emsg import *

if TYPE_CHECKING:
    from collections.abc import Mapping

    import betterproto

    from ..enums import IntEnum

__all__ = ("PROTOBUFS",)


PROTOBUFS: Mapping[IntEnum, type[betterproto.Message]] = {
    EMsg.Multi: base.CMsgMulti,
    EMsg.ClientToGC: client_server_2.CMsgGcClient,
    EMsg.ClientFromGC: client_server_2.CMsgGcClient,
    EMsg.ClientGetNumberOfCurrentPlayersDP: client_server_2.CMsgDpGetNumberOfCurrentPlayers,
    EMsg.ClientGetNumberOfCurrentPlayersDPResponse: client_server_2.CMsgDpGetNumberOfCurrentPlayersResponse,
    EMsg.ClientLogonGameServer: login.CMsgClientLogon,
    EMsg.ClientCurrentUIMode: client_server_2.CMsgClientUiMode,
    EMsg.ClientChatOfflineMessageNotification: client_server_2.CMsgClientOfflineMessageNotification,
    EMsg.ClientTicketAuthComplete: client_server.CMsgClientTicketAuthComplete,
    EMsg.ClientCMList: client_server.CMsgClientCmList,
    EMsg.ClientP2PConnectionInfo: client_server.CMsgClientP2PConnectionInfo,
    EMsg.ClientP2PConnectionFailInfo: client_server.CMsgClientP2PConnectionFailInfo,
    EMsg.ClientNetworkingCertRequest: client_server.CMsgClientNetworkingCertRequest,
    EMsg.ClientNetworkingMobileCertRequest: client_server.CMsgClientNetworkingMobileCertRequest,
    EMsg.ClientGetAppOwnershipTicket: client_server.CMsgClientGetAppOwnershipTicket,
    EMsg.ClientGetAppOwnershipTicketResponse: client_server.CMsgClientGetAppOwnershipTicketResponse,
    EMsg.ClientSessionToken: client_server.CMsgClientSessionToken,
    EMsg.ClientGameConnectTokens: client_server.CMsgClientGameConnectTokens,
    EMsg.ClientGamesPlayed: client_server.CMsgClientGamesPlayed,
    EMsg.GSApprove: client_server.CMsgGsApprove,
    EMsg.GSDeny: client_server.CMsgGsDeny,
    EMsg.GSKick: client_server.CMsgGsKick,
    EMsg.ClientAuthList: client_server.CMsgClientAuthList,
    EMsg.ClientAuthListAck: client_server.CMsgClientAuthListAck,
    EMsg.ClientLicenseList: client_server.CMsgClientLicenseList,
    EMsg.ClientGamesPlayedWithDataBlob: client_server.CMsgClientGamesPlayed,
    EMsg.ClientIsLimitedAccount: client_server.CMsgClientIsLimitedAccount,
    EMsg.ClientRequestedClientStats: client_server.CMsgClientRequestedClientStats,
    EMsg.ClientStat2: client_server.CMsgClientStat2,
    EMsg.ClientInviteToGame: client_server.CMsgClientInviteToGame,
    EMsg.ClientChatInvite: client_server.CMsgClientChatInvite,
    EMsg.ClientConnectionStats: client_server.CMsgClientConnectionStats,
    EMsg.ClientServersAvailable: client_server.CMsgClientServersAvailable,
    EMsg.ClientGetUserStats: user_stats.CMsgClientGetUserStats,
    EMsg.ClientGetUserStatsResponse: user_stats.CMsgClientGetUserStatsResponse,
    EMsg.ClientStoreUserStatsResponse: user_stats.CMsgClientStoreUserStatsResponse,
    EMsg.ClientStoreUserStats2: user_stats.CMsgClientStoreUserStats2,
    EMsg.ClientStatsUpdated: user_stats.CMsgClientStatsUpdated,
    EMsg.ClientStoreUserStats: user_stats.CMsgClientStoreUserStats,
    EMsg.ClientReportOverlayDetourFailure: client_server.CMsgClientReportOverlayDetourFailure,
    EMsg.ClientWalletInfoUpdate: client_server.CMsgClientWalletInfoUpdate,
    EMsg.ClientPICSChangesSinceRequest: app_info.CMsgClientPicsChangesSinceRequest,
    EMsg.ClientPICSChangesSinceResponse: app_info.CMsgClientPicsChangesSinceResponse,
    EMsg.ClientPICSProductInfoRequest: app_info.CMsgClientPicsProductInfoRequest,
    EMsg.ClientPICSProductInfoResponse: app_info.CMsgClientPicsProductInfoResponse,
    EMsg.ClientPICSAccessTokenRequest: app_info.CMsgClientPicsAccessTokenRequest,
    EMsg.ClientPICSAccessTokenResponse: app_info.CMsgClientPicsAccessTokenResponse,
    EMsg.ClientDeregisterWithServer: client_server.CMsgClientDeregisterWithServer,
    EMsg.ClientClanState: client_server.CMsgClientClanState,
    EMsg.ClientScreenshotsChanged: ucm.CMsgClientScreenshotsChanged,
    EMsg.ClientUpdateUserGameInfo: client_server_2.CMsgClientUpdateUserGameInfo,
    EMsg.ClientRichPresenceUpload: client_server_2.CMsgClientRichPresenceUpload,
    EMsg.ClientRichPresenceRequest: client_server_2.CMsgClientRichPresenceRequest,
    EMsg.ClientRichPresenceInfo: client_server_2.CMsgClientRichPresenceInfo,
    EMsg.ClientReadMachineAuth: client_server_2.CMsgClientReadMachineAuth,
    EMsg.ClientReadMachineAuthResponse: client_server_2.CMsgClientReadMachineAuthResponse,
    EMsg.ClientUpdateMachineAuthResponse: client_server_2.CMsgClientUpdateMachineAuthResponse,
    EMsg.ClientRequestMachineAuth: client_server_2.CMsgClientRequestMachineAuth,
    EMsg.ClientRequestMachineAuthResponse: client_server_2.CMsgClientRequestMachineAuthResponse,
    EMsg.ClientRegisterKey: client_server_2.CMsgClientRegisterKey,
    EMsg.ClientPurchaseResponse: client_server_2.CMsgClientPurchaseResponse,
    EMsg.ClientActivateOEMLicense: client_server_2.CMsgClientActivateOemLicense,
    EMsg.ClientRegisterOEMMachine: client_server_2.CMsgClientRegisterOemMachine,
    EMsg.ClientRegisterOEMMachineResponse: client_server_2.CMsgClientRegisterOemMachineResponse,
    EMsg.ClientPurchaseWithMachineID: client_server_2.CMsgClientPurchaseWithMachineId,
    EMsg.ClientGetCDNAuthToken: client_server_2.CMsgClientGetCdnAuthToken,
    EMsg.ClientGetDepotDecryptionKey: client_server_2.CMsgClientGetDepotDecryptionKey,
    EMsg.ClientGetDepotDecryptionKeyResponse: client_server_2.CMsgClientGetDepotDecryptionKeyResponse,
    EMsg.ClientCheckAppBetaPassword: client_server_2.CMsgClientCheckAppBetaPassword,
    EMsg.ClientCheckAppBetaPasswordResponse: client_server_2.CMsgClientCheckAppBetaPasswordResponse,
    EMsg.ClientDPContentStatsReport: client_server_2.CMsgClientDpContentStatsReport,
    EMsg.ClientGetCDNAuthTokenResponse: client_server_2.CMsgClientGetCdnAuthTokenResponse,
    EMsg.ClientRequestAccountData: client_server_2.CMsgClientRequestAccountData,
    EMsg.ClientRequestAccountDataResponse: client_server_2.CMsgClientRequestAccountDataResponse,
    EMsg.ClientRedeemGuestPass: client_server_2.CMsgClientRedeemGuestPass,
    EMsg.ClientRedeemGuestPassResponse: client_server_2.CMsgClientRedeemGuestPassResponse,
    EMsg.ClientGetClanActivityCounts: client_server_2.CMsgClientGetClanActivityCounts,
    EMsg.ClientGetClanActivityCountsResponse: client_server_2.CMsgClientGetClanActivityCountsResponse,
    EMsg.ClientOGSReportString: client_server_2.CMsgClientOgsReportString,
    EMsg.ClientOGSReportBug: client_server_2.CMsgClientOgsReportBug,
    EMsg.ClientSentLogs: client_server_2.CMsgClientSentLogs,
    EMsg.ClientRequestFreeLicense: client_server_2.CMsgClientRequestFreeLicense,
    EMsg.ClientRequestFreeLicenseResponse: client_server_2.CMsgClientRequestFreeLicenseResponse,
    EMsg.ClientRequestForgottenPasswordEmail: client_server_2.CMsgClientRequestForgottenPasswordEmail,
    EMsg.ClientRequestForgottenPasswordEmailResponse: client_server_2.CMsgClientRequestForgottenPasswordEmailResponse,
    EMsg.ClientItemAnnouncements: client_server_2.CMsgClientItemAnnouncements,
    EMsg.ClientRequestItemAnnouncements: client_server_2.CMsgClientRequestItemAnnouncements,
    EMsg.ClientUserNotifications: client_server_2.CMsgClientUserNotifications,
    EMsg.ClientCommentNotifications: client_server_2.CMsgClientCommentNotifications,
    EMsg.ClientRequestCommentNotifications: client_server_2.CMsgClientRequestCommentNotifications,
    EMsg.ClientChatGetFriendMessageHistory: client_server_2.CMsgClientChatGetFriendMessageHistory,
    EMsg.ClientChatGetFriendMessageHistoryResponse: client_server_2.CMsgClientChatGetFriendMessageHistoryResponse,
    EMsg.ClientChatGetFriendMessageHistoryForOfflineMessages: client_server_2.CMsgClientChatGetFriendMessageHistoryForOfflineMessages,
    EMsg.ClientFSGetFriendsSteamLevels: client_server_2.CMsgClientFsGetFriendsSteamLevels,
    EMsg.ClientFSGetFriendsSteamLevelsResponse: client_server_2.CMsgClientFsGetFriendsSteamLevelsResponse,
    EMsg.ClientEmailAddrInfo: client_server_2.CMsgClientEmailAddrInfo,
    EMsg.ClientFriendUserStatusPublished: client_server_2.CMsgClientFriendUserStatusPublished,
    EMsg.ClientServiceMethodLegacy: client_server_2.CMsgClientServiceMethodLegacy,
    EMsg.ClientServiceMethodLegacyResponse: client_server_2.CMsgClientServiceMethodLegacyResponse,
    EMsg.ClientAuthorizeLocalDeviceRequest: client_server_2.CMsgClientAuthorizeLocalDeviceRequest,
    EMsg.ClientAuthorizeLocalDeviceNotification: client_server_2.CMsgClientAuthorizeLocalDeviceNotification,
    EMsg.ClientDeauthorizeDeviceRequest: client_server_2.CMsgClientDeauthorizeDeviceRequest,
    EMsg.ClientDeauthorizeDevice: client_server_2.CMsgClientDeauthorizeDevice,
    EMsg.ClientUseLocalDeviceAuthorizations: client_server_2.CMsgClientUseLocalDeviceAuthorizations,
    EMsg.ClientGetAuthorizedDevices: client_server_2.CMsgClientGetAuthorizedDevices,
    EMsg.ClientGetAuthorizedDevicesResponse: client_server_2.CMsgClientGetAuthorizedDevicesResponse,
    EMsg.ClientSharedLibraryLockStatus: client_server_2.CMsgClientSharedLibraryLockStatus,
    EMsg.ClientSharedLibraryStopPlaying: client_server_2.CMsgClientSharedLibraryStopPlaying,
    EMsg.ClientServiceCall: client_server_2.CMsgClientServiceCall,
    EMsg.ClientServiceModule: client_server_2.CMsgClientServiceModule,
    EMsg.ClientServiceCallResponse: client_server_2.CMsgClientServiceCallResponse,
    EMsg.ClientPlayingSessionState: client_server_2.CMsgClientPlayingSessionState,
    EMsg.ClientKickPlayingSession: client_server_2.CMsgClientKickPlayingSession,
    EMsg.ClientVoiceCallPreAuthorizeResponse: client_server_2.CMsgClientVoiceCallPreAuthorizeResponse,
    EMsg.BadgeCraftedNotification: client_server_2.CMsgBadgeCraftedNotification,
    EMsg.ClientFriendMsg: friends.CMsgClientFriendMsg,
    EMsg.ClientFriendMsgIncoming: friends.CMsgClientFriendMsgIncoming,
    EMsg.ClientAddFriend: friends.CMsgClientAddFriend,
    EMsg.ClientAddFriendResponse: friends.CMsgClientAddFriendResponse,
    EMsg.ClientRemoveFriend: friends.CMsgClientRemoveFriend,
    EMsg.ClientHideFriend: friends.CMsgClientHideFriend,
    EMsg.ClientFriendsList: friends.CMsgClientFriendsList,
    EMsg.ClientFriendsGroupsList: friends.CMsgClientFriendsGroupsList,
    EMsg.ClientRequestFriendData: friends.CMsgClientRequestFriendData,
    EMsg.ClientChangeStatus: friends.CMsgClientChangeStatus,
    EMsg.ClientPersonaState: friends.CMsgClientPersonaState,
    EMsg.ClientFriendProfileInfo: friends.CMsgClientFriendProfileInfo,
    EMsg.ClientFriendProfileInfoResponse: friends.CMsgClientFriendProfileInfoResponse,
    EMsg.ClientGetEmoticonList: friends.CMsgClientGetEmoticonList,
    EMsg.ClientEmoticonList: friends.CMsgClientEmoticonList,
    EMsg.ClientHeartBeat: login.CMsgClientHeartBeat,
    EMsg.ClientServerTimestampRequest: login.CMsgClientServerTimestampRequest,
    EMsg.ClientServerTimestampResponse: login.CMsgClientServerTimestampResponse,
    EMsg.ClientLogon: login.CMsgClientLogon,
    EMsg.ClientLogOnResponse: login.CMsgClientLogonResponse,
    EMsg.ClientLogOff: login.CMsgClientLogOff,
    EMsg.ClientLoggedOff: login.CMsgClientLoggedOff,
    EMsg.ClientNewLoginKey: login.CMsgClientNewLoginKey,
    EMsg.ClientNewLoginKeyAccepted: login.CMsgClientNewLoginKeyAccepted,
    EMsg.ClientAccountInfo: login.CMsgClientAccountInfo,
    EMsg.ClientChallengeRequest: login.CMsgClientChallengeRequest,
    EMsg.ClientChallengeResponse: login.CMsgClientChallengeResponse,
    EMsg.ClientLBSFindOrCreateLB: leaderboards.CMsgClientLbsFindOrCreateLb,
    EMsg.ClientLBSFindOrCreateLBResponse: leaderboards.CMsgClientLbsFindOrCreateLbResponse,
    EMsg.ClientLBSGetLBEntries: leaderboards.CMsgClientLbsGetLbEntries,
    EMsg.ClientLBSGetLBEntriesResponse: leaderboards.CMsgClientLbsGetLbEntriesResponse,
}



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING

from . import utils
from .enums import ProfileItemType, Result
from .errors import WSException
from .game import StatefulGame

if TYPE_CHECKING:
    from .protobufs import player
    from .state import ConnectionState

__all__ = (
    "ProfileInfo",
    "ProfileItem",
    "OwnedProfileItems",
    "EquippedProfileItems",
    "Profile",
)


@dataclass
class ProfileInfo:
    """Represents the user's profile info.

    Attributes
    ----------
    created_at
        The time at which the user was created at.
    real_name
        The real name of the user.
    city_name
        The city the user is located in.
    state_name
        The name of the state the user is located in.
    country_name
        The name of the country the user is located in.
    headline
        The profile's headline.
    summary
        The user's summary.
    """

    created_at: datetime
    real_name: str | None
    city_name: str | None
    state_name: str | None
    country_name: str | None
    headline: str | None
    summary: str


@dataclass
class ProfileMovie:
    url: str  # TODO add more attributes like maybe created_at?


class ProfileItem:
    """Represents an item on/in a user's profile.

    Attributes
    ----------
    id
        The item's id.
    url
        The item's url.
    name
        The item's name.
    title
        The item's title.
    description
        The item's description.
    game
        The game the item is from.
    type
        The item's type.
    class_
        The item's class.
    movie
        The movie associated with the item.
    equipped_flags
        The item's equipped flags.
    """

    __slots__ = (
        "id",
        "url",
        "name",
        "title",
        "description",
        "game",
        "type",
        "class_",
        "movie",
        "equipped_flags",
        "_state",
        "_um_name",
    )

    def __init__(self, state: ConnectionState, item: player.ProfileItem, *, um_name: str | None = None):
        self.id = item.communityitemid
        self.url = item.image_large
        self.name = item.name
        self.title = item.item_title
        self.description = item.item_description
        self.game = StatefulGame(state, id=item.appid)
        self.type = ProfileItemType.try_value(item.item_type)
        self.class_ = item.item_class
        self.movie = ProfileMovie(item.movie_mp4)
        self.equipped_flags = item.equipped_flags  # TODO might be useful for item show case?
        self._state = state
        self._um_name = um_name

    def __repr__(self) -> str:
        return f"<ProfileItem id={self.id} name={self.name!r} game={self.game!r}>"

    async def equip(self) -> None:
        """Equip the profile item."""
        if self._um_name is None:
            raise ValueError(f"Cannot equip {self!r}")

        msg = await self._state.ws.send_um_and_wait(f"Player.Set{self._um_name}", communityitemid=self.id)
        if msg.result != Result.OK:
            raise WSException(msg)


@dataclass
class OwnedProfileItems:
    r"""Represents the :class:`ClientUser`\'s owned items.

    Attributes
    ----------
    backgrounds
        The backgrounds the client user owns.
    mini_profile_backgrounds
        The mini profile backgrounds the client user owns.
    avatar_frames
        The avatar frames the client user owns.
    animated_avatars
        The animated avatars the client user owns.
    modifiers
        The modifiers the client user owns.
    """
    __slots__ = ("backgrounds", "mini_profile_backgrounds", "avatar_frames", "animated_avatars", "modifiers")
    backgrounds: list[ProfileItem]
    mini_profile_backgrounds: list[ProfileItem]
    avatar_frames: list[ProfileItem]
    animated_avatars: list[ProfileItem]
    modifiers: list[ProfileItem]


@dataclass
class EquippedProfileItems:
    """Represents the items the user has equipped.

    Attributes
    ----------
    background
        The equipped background.
    mini_profile_background
        The equipped mini profile background for the user.
    avatar_frame
        The equipped avatar frame for the user.
    animated_avatar
        The equipped animated avatar for the user.
    modifier
        The equipped modifier for the user.
    """

    background: ProfileItem | None
    mini_profile_background: ProfileItem | None
    avatar_frame: ProfileItem | None
    animated_avatar: ProfileItem | None
    modifier: ProfileItem | None


class Profile(ProfileInfo, EquippedProfileItems):
    r"""Represents a user's complete profile.

    Attributes
    ----------
    background
        The equipped background.
    mini_profile_background
        The equipped mini profile background for the user.
    avatar_frame
        The equipped avatar frame for the user.
    animated_avatar
        The equipped animated avatar for the user.
    modifier
        The equipped modifier for the user.
    items
        The account's owned profile items.

        Note
        ----
        This is only available for the :class:`ClientUser`\'s profile otherwise it is ``None``.
    """

    def __init__(self, equipped_items: EquippedProfileItems, info: ProfileInfo, items: OwnedProfileItems | None = None):
        utils.update_class(equipped_items, self)
        utils.update_class(info, self)
        self.items = items

    def __repr__(self) -> str:
        return f"<Profile real_name={self.real_name!r}>"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: encrypted_app_ticket.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass

import betterproto


@dataclass(eq=False, repr=False)
class EncryptedAppTicket(betterproto.Message):
    ticket_version_no: int = betterproto.uint32_field(1)
    crc_encryptedticket: int = betterproto.uint32_field(2)
    cb_encrypteduserdata: int = betterproto.uint32_field(3)
    cb_encrypted_appownershipticket: int = betterproto.uint32_field(4)
    encrypted_ticket: bytes = betterproto.bytes_field(5)



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from collections import deque
from typing import Dict, Generator, TypeVar, overload

from .errors import MissingClosingQuotation

_T = TypeVar("_T")
_VT = TypeVar("_VT")


class CaseInsensitiveDict(Dict[str, _VT]):
    """A dictionary where keys are case insensitive."""

    def __init__(self, **kwargs: _VT):
        super().__init__(**{k.lower(): v for k, v in kwargs.items()})

    def __repr__(self) -> str:
        return f"CaseInsensitiveDict({', '.join(f'{k}={v!r}' for k, v in self.items())})"

    def __contains__(self, k: str) -> bool:
        return super().__contains__(k.lower())

    def __delitem__(self, k: str) -> None:
        super().__delitem__(k.lower())

    def __getitem__(self, k: str) -> _VT:
        return super().__getitem__(k.lower())

    def __setitem__(self, k: str, v: _VT) -> None:
        super().__setitem__(k.lower(), v)

    @overload
    def get(self, k: str) -> _VT | None:
        ...

    @overload
    def get(self, k: str, default: _T | None = None) -> _VT | _T | None:
        ...

    def get(self, k: str, default: _T | None = None) -> _VT | _T | None:
        return super().get(k.lower(), default)

    @overload
    def pop(self, k: str) -> _VT | None:
        ...

    @overload
    def pop(self, k: str, default: _T | None = None) -> _VT | _T | None:
        ...

    def pop(self, k: str, default: _T | None = None) -> _VT | _T | None:
        return super().pop(k.lower())


def _end_of_quote_finder(in_stream: str, location: int) -> int:
    end_of_quote_index = in_stream.find('"', location)
    if end_of_quote_index == -1:
        raise MissingClosingQuotation(location)
    if in_stream[end_of_quote_index - 1] == "\\":  # quote is escaped carry on searching
        return _end_of_quote_finder(in_stream, end_of_quote_index + 1)
    return end_of_quote_index


def remove_quotes(string: str) -> str:
    """
    >>> remove_quotes('"only a quote at the start')
    '"quoted only at the start'
    >>> remove_quotes('"quoted all the way"')
    'quoted all the way'
    """
    return string[1:-1] if (string[-1:], string[:1]) == ('"', '"') else string


class Shlex:
    """A simple lexical analyser.
    This should be a simpler and faster version of :class:`shlex.shlex` in posix mode.
    """

    __slots__ = ("in_stream", "position", "end", "_undo_pushback")

    def __init__(self, in_stream: str):
        self.in_stream = in_stream.replace("‘", '"').replace("’", '"').replace("“", '"').replace("”", '"').strip()
        self.position = 0
        self.end = len(self.in_stream)
        self._undo_pushback: deque[int] = deque()

    def read(self) -> str | None:
        if self.position >= self.end:
            return None

        while True:
            start = self.position
            characters = []

            for character in self.in_stream[self.position :]:
                self.position += 1
                if character.isspace():
                    break

                if character == '"':
                    before = self.position
                    if self.in_stream[before - 2] != "\\":  # quote is escaped carry on searching
                        end_of_quote = _end_of_quote_finder(self.in_stream, before)
                        self.position = end_of_quote + 2
                        self._undo_pushback.append(start)
                        return remove_quotes(self.in_stream[before:end_of_quote])

                    characters.pop()
                characters.append(character)

            ret = remove_quotes("".join(characters))
            if ret:
                self._undo_pushback.append(start)
                return ret

    def undo(self) -> None:
        try:
            self.position = self._undo_pushback.pop()
        except IndexError:
            pass

    @property
    def rest(self) -> str:
        return self.in_stream[self.position :]

    def __repr__(self) -> str:
        attrs = (
            "in_stream",
            "position",
            "end",
        )
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Shlex {' '.join(resolved)}>"

    def __iter__(self) -> Generator[str, None, None]:
        while True:
            token = self.read()
            if token is None:
                break
            yield token



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING, Generic, TypeVar, Union

from ...enums import IntEnum
from .errors import CommandOnCooldown

if TYPE_CHECKING:
    from ...abc import Message, Messageable

__all__ = (
    "BucketType",
    "Cooldown",
)
T_Bucket = TypeVar("T_Bucket", "tuple[int, ...]", int)


class BucketType(IntEnum):
    """A enumeration that handles cooldowns.

    Each attribute operates on a per x basis. So :attr:`User` operates on a per :class:`~steam.User` basis.
    """

    # fmt: off
    Default = 0  #: The default :class:`BucketType` this operates on a global basis.
    User    = 1  #: The :class:`BucketType` for a :class:`steam.User`.
    Member  = 2  #: The :class:`BucketType` for a :class:`steam.User` and a :class:`steam.Clan`.
    Group   = 3  #: The :class:`BucketType` for a :class:`steam.User` in a :class:`steam.Clan` / :class:`steam.Group`.
    Clan    = 4  #: The :class:`BucketType` for a :class:`steam.Clan`.
    Role    = 5  #: The :class:`BucketType` for a :class:`steam.Role`.
    Channel = 6  #: The :class:`BucketType` for a :class:`steam.Channel`.
    Admin   = 7  #: The :class:`BucketType` for a :class:`steam.Clan`'s :attr:`steam.Clan.admins`.
    # fmt: on

    def get_bucket(self, ctx: Message | Messageable) -> int | tuple[int, ...]:
        """Get a bucket key for a message or context.

        Parameters
        ----------
        ctx
            The message or context to get the bucket for.
        """
        if self == BucketType.Default:
            return 0
        elif self == BucketType.User:
            return ctx.author.id
        elif self == BucketType.Member:
            return (ctx.clan and ctx.clan.id), ctx.author.id
        elif self == BucketType.Group:
            return (ctx.group or ctx.author).id
        elif self == BucketType.Role:
            return (ctx.clan and ctx.author.top_role.id), ctx.author.id
        elif self == BucketType.Clan:
            return (ctx.clan or ctx.author).id
        elif self == BucketType.Channel:
            return (ctx.channel or ctx.author).id
        elif self == BucketType.Admin:
            return (ctx.clan and (ctx.author in ctx.clan.admins)), ctx.author.id


class Cooldown(Generic[T_Bucket]):
    """The class that holds a command's cooldown."""

    bucket: BucketType  #: The bucket that should be used to determine this command's cooldown.

    def __init__(self, rate: int, per: float, bucket: BucketType):
        self._rate = rate
        self._per = per
        self.bucket = bucket
        self.reset()

    def reset(self) -> None:
        """Reset the command's cooldown."""
        self._queue: dict[float, T_Bucket] = {}

    def _calls(self, bucket: T_Bucket) -> list[tuple[float, T_Bucket], ...]:
        return [(t, b) for t, b in self._queue.items() if b == bucket]

    async def expire_cache(self, bucket: T_Bucket, now: float) -> None:
        self._queue[now] = bucket
        await asyncio.sleep(self._per)
        self._queue.pop(now, None)

    def get_retry_after(self, bucket: T_Bucket, now: float) -> float:
        """Get the retry after for a command.

        Parameters
        ----------
        bucket
            The bucket to find in the cache.
        now
            The UNIX timestamp to find times after.
        """
        calls = self._calls(bucket)
        if not calls:
            return 0.0
        last_call = calls[0][0]
        if last_call + self._per >= now and len(calls) >= self._rate:
            return last_call + self._per - now
        return 0.0

    def __call__(self, ctx: Message | Messageable) -> None:
        """Invoke the command's cooldown properly and raise if the command is on cooldown.

        Parameters
        ----------
        ctx
            The context for invocation to check for a cooldown on.
        """
        bucket = self.bucket.get_bucket(ctx)
        now = time.time()
        asyncio.create_task(self.expire_cache(bucket, now))
        retry_after = self.get_retry_after(bucket, now)
        if retry_after:
            raise CommandOnCooldown(retry_after)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_base.proto
# plugin: python-betterproto
# Last updated 09/09/2021
from dataclasses import dataclass
from typing import List

import betterproto


class EProtoClanEventType(betterproto.Enum):
    OtherEvent = 1
    GameEvent = 2
    PartyEvent = 3
    MeetingEvent = 4
    SpecialCauseEvent = 5
    MusicAndArtsEvent = 6
    SportsEvent = 7
    TripEvent = 8
    ChatEvent = 9
    GameReleaseEvent = 10
    BroadcastEvent = 11
    SmallUpdateEvent = 12
    PreAnnounceMajorUpdateEvent = 13
    MajorUpdateEvent = 14
    DLCReleaseEvent = 15
    FutureReleaseEvent = 16
    ESportTournamentStreamEvent = 17
    DevStreamEvent = 18
    FamousStreamEvent = 19
    GameSalesEvent = 20
    GameItemSalesEvent = 21
    InGameBonusXPEvent = 22
    InGameLootEvent = 23
    InGamePerksEvent = 24
    InGameChallengeEvent = 25
    InGameContestEvent = 26
    IRLEvent = 27
    NewsEvent = 28
    BetaReleaseEvent = 29
    InGameContentReleaseEvent = 30
    FreeTrial = 31
    SeasonRelease = 32
    SeasonUpdate = 33
    CrosspostEvent = 34
    InGameEventGeneral = 35


@dataclass(eq=False, repr=False)
class CMsgIpAddress(betterproto.Message):
    v4: int = betterproto.fixed32_field(1, group="ip")
    v6: bytes = betterproto.bytes_field(2, group="ip")


@dataclass(eq=False, repr=False)
class CMsgIpAddressBucket(betterproto.Message):
    original_ip_address: "CMsgIpAddress" = betterproto.message_field(1)
    bucket: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class CMsgProtoBufHeader(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    session_id: int = betterproto.int32_field(2)
    routing_app_id: int = betterproto.uint32_field(3)
    job_id_source: int = betterproto.fixed64_field(10)
    job_id_target: int = betterproto.fixed64_field(11)
    job_name_target: str = betterproto.string_field(12)
    seq_num: int = betterproto.int32_field(24)
    eresult: int = betterproto.int32_field(13)
    error_message: str = betterproto.string_field(14)
    auth_account_flags: int = betterproto.uint32_field(16)
    token_source: int = betterproto.uint32_field(22)
    admin_spoofing_user: bool = betterproto.bool_field(23)
    transport_error: int = betterproto.int32_field(17)
    message_id: int = betterproto.uint64_field(18)
    publisher_group_id: int = betterproto.uint32_field(19)
    sys_id: int = betterproto.uint32_field(20)
    trace_tag: int = betterproto.uint64_field(21)
    webapi_key_id: int = betterproto.uint32_field(25)
    is_from_external_source: bool = betterproto.bool_field(26)
    forward_to_sys_id: List[int] = betterproto.uint32_field(27)
    cm_sys_id: int = betterproto.uint32_field(28)
    wg_token: str = betterproto.string_field(30)
    launcher_type: int = betterproto.uint32_field(31)
    realm: int = betterproto.uint32_field(32)
    ip: int = betterproto.uint32_field(15, group="ip_addr")
    ip_v6: bytes = betterproto.bytes_field(29, group="ip_addr")


@dataclass(eq=False, repr=False)
class CMsgMulti(betterproto.Message):
    size_unzipped: int = betterproto.uint32_field(1)
    message_body: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgProtobufWrapped(betterproto.Message):
    message_body: bytes = betterproto.bytes_field(1)


@dataclass(eq=False, repr=False)
class CMsgAuthTicket(betterproto.Message):
    estate: int = betterproto.uint32_field(1)
    eresult: int = betterproto.uint32_field(2)
    steamid: int = betterproto.fixed64_field(3)
    gameid: int = betterproto.fixed64_field(4)
    h_steam_pipe: int = betterproto.uint32_field(5)
    ticket_crc: int = betterproto.uint32_field(6)
    ticket: bytes = betterproto.bytes_field(7)


@dataclass(eq=False, repr=False)
class CcddbAppDetailCommon(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    name: str = betterproto.string_field(2)
    icon: str = betterproto.string_field(3)
    logo: str = betterproto.string_field(4)
    logo_small: str = betterproto.string_field(5)
    tool: bool = betterproto.bool_field(6)
    demo: bool = betterproto.bool_field(7)
    media: bool = betterproto.bool_field(8)
    community_visible_stats: bool = betterproto.bool_field(9)
    friendly_name: str = betterproto.string_field(10)
    propagation: str = betterproto.string_field(11)
    has_adult_content: bool = betterproto.bool_field(12)
    is_visible_in_steam_china: bool = betterproto.bool_field(13)


@dataclass(eq=False, repr=False)
class CMsgAppRights(betterproto.Message):
    edit_info: bool = betterproto.bool_field(1)
    publish: bool = betterproto.bool_field(2)
    view_error_data: bool = betterproto.bool_field(3)
    download: bool = betterproto.bool_field(4)
    upload_cdkeys: bool = betterproto.bool_field(5)
    generate_cdkeys: bool = betterproto.bool_field(6)
    view_financials: bool = betterproto.bool_field(7)
    manage_ceg: bool = betterproto.bool_field(8)
    manage_signing: bool = betterproto.bool_field(9)
    manage_cdkeys: bool = betterproto.bool_field(10)
    edit_marketing: bool = betterproto.bool_field(11)
    economy_support: bool = betterproto.bool_field(12)
    economy_support_supervisor: bool = betterproto.bool_field(13)
    manage_pricing: bool = betterproto.bool_field(14)
    broadcast_live: bool = betterproto.bool_field(15)
    view_marketing_traffic: bool = betterproto.bool_field(16)
    edit_store_display_content: bool = betterproto.bool_field(17)


@dataclass(eq=False, repr=False)
class CCuratorPreferences(betterproto.Message):
    supported_languages: int = betterproto.uint32_field(1)
    platform_windows: bool = betterproto.bool_field(2)
    platform_mac: bool = betterproto.bool_field(3)
    platform_linux: bool = betterproto.bool_field(4)
    vr_content: bool = betterproto.bool_field(5)
    adult_content_violence: bool = betterproto.bool_field(6)
    adult_content_sex: bool = betterproto.bool_field(7)
    timestamp_updated: int = betterproto.uint32_field(8)
    tagids_curated: List[int] = betterproto.uint32_field(9)
    tagids_filtered: List[int] = betterproto.uint32_field(10)
    website_title: str = betterproto.string_field(11)
    website_url: str = betterproto.string_field(12)
    discussion_url: str = betterproto.string_field(13)
    show_broadcast: bool = betterproto.bool_field(14)


@dataclass(eq=False, repr=False)
class CLocalizationToken(betterproto.Message):
    language: int = betterproto.uint32_field(1)
    localized_string: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CClanEventUserNewsTuple(betterproto.Message):
    clanid: int = betterproto.uint32_field(1)
    event_gid: int = betterproto.fixed64_field(2)
    announcement_gid: int = betterproto.fixed64_field(3)
    rtime_start: int = betterproto.uint32_field(4)
    rtime_end: int = betterproto.uint32_field(5)
    priority_score: int = betterproto.uint32_field(6)
    type: int = betterproto.uint32_field(7)
    clamp_range_slot: int = betterproto.uint32_field(8)
    appid: int = betterproto.uint32_field(9)
    rtime32_last_modified: int = betterproto.uint32_field(10)


@dataclass(eq=False, repr=False)
class CClanMatchEventByRange(betterproto.Message):
    rtime_before: int = betterproto.uint32_field(1)
    rtime_after: int = betterproto.uint32_field(2)
    qualified: int = betterproto.uint32_field(3)
    events: List["CClanEventUserNewsTuple"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class CCommunityClanAnnouncementInfo(betterproto.Message):
    gid: int = betterproto.uint64_field(1)
    clanid: int = betterproto.uint64_field(2)
    posterid: int = betterproto.uint64_field(3)
    headline: str = betterproto.string_field(4)
    posttime: int = betterproto.uint32_field(5)
    updatetime: int = betterproto.uint32_field(6)
    body: str = betterproto.string_field(7)
    commentcount: int = betterproto.int32_field(8)
    tags: List[str] = betterproto.string_field(9)
    language: int = betterproto.int32_field(10)
    hidden: bool = betterproto.bool_field(11)
    forum_topic_id: int = betterproto.fixed64_field(12)
    event_gid: int = betterproto.fixed64_field(13)
    voteupcount: int = betterproto.int32_field(14)
    votedowncount: int = betterproto.int32_field(15)


@dataclass(eq=False, repr=False)
class CClanEventData(betterproto.Message):
    gid: int = betterproto.fixed64_field(1)
    clan_steamid: int = betterproto.fixed64_field(2)
    event_name: str = betterproto.string_field(3)
    event_type: "EProtoClanEventType" = betterproto.enum_field(4)
    appid: int = betterproto.uint32_field(5)
    server_address: str = betterproto.string_field(6)
    server_password: str = betterproto.string_field(7)
    rtime32_start_time: int = betterproto.uint32_field(8)
    rtime32_end_time: int = betterproto.uint32_field(9)
    comment_count: int = betterproto.int32_field(10)
    creator_steamid: int = betterproto.fixed64_field(11)
    last_update_steamid: int = betterproto.fixed64_field(12)
    event_notes: str = betterproto.string_field(13)
    jsondata: str = betterproto.string_field(14)
    announcement_body: "CCommunityClanAnnouncementInfo" = betterproto.message_field(15)
    published: bool = betterproto.bool_field(16)
    hidden: bool = betterproto.bool_field(17)
    rtime32_visibility_start: int = betterproto.uint32_field(18)
    rtime32_visibility_end: int = betterproto.uint32_field(19)
    broadcaster_accountid: int = betterproto.uint32_field(20)
    follower_count: int = betterproto.uint32_field(21)
    ignore_count: int = betterproto.uint32_field(22)
    forum_topic_id: int = betterproto.fixed64_field(23)
    rtime32_last_modified: int = betterproto.uint32_field(24)
    news_post_gid: int = betterproto.fixed64_field(25)
    rtime_mod_reviewed: int = betterproto.uint32_field(26)
    featured_app_tagid: int = betterproto.uint32_field(27)
    referenced_appids: List[int] = betterproto.uint32_field(28)
    build_id: int = betterproto.uint32_field(29)
    build_branch: str = betterproto.string_field(30)


@dataclass(eq=False, repr=False)
class CBillingAddress(betterproto.Message):
    first_name: str = betterproto.string_field(1)
    last_name: str = betterproto.string_field(2)
    address1: str = betterproto.string_field(3)
    address2: str = betterproto.string_field(4)
    city: str = betterproto.string_field(5)
    us_state: str = betterproto.string_field(6)
    country_code: str = betterproto.string_field(7)
    postcode: str = betterproto.string_field(8)
    zip_plus4: int = betterproto.int32_field(9)
    phone: str = betterproto.string_field(10)


@dataclass(eq=False, repr=False)
class CPackageReservationStatus(betterproto.Message):
    packageid: int = betterproto.uint32_field(1)
    reservation_state: int = betterproto.int32_field(2)
    queue_position: int = betterproto.int32_field(3)
    total_queue_size: int = betterproto.int32_field(4)
    reservation_country_code: str = betterproto.string_field(5)
    expired: bool = betterproto.bool_field(6)
    time_expires: int = betterproto.uint32_field(7)
    time_reserved: int = betterproto.uint32_field(8)



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import dis
import sys
import types
from collections.abc import Iterator, Sequence
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeVar

from typing_extensions import NotRequired, Required, TypeAlias, TypedDict

from . import utils
from .enums import TradeOfferState
from .errors import ClientException, ConfirmationError
from .game import Game, StatefulGame

if TYPE_CHECKING:
    from .abc import BaseUser, SteamID
    from .state import ConnectionState
    from .user import User


__all__ = (
    "Item",
    "Asset",
    "Inventory",
    "TradeOffer",
    "TradeOfferReceiptItem",
    "TradeOfferReceipt",
)

Items: TypeAlias = "Item | Asset"
I = TypeVar("I", bound="Item")
T = TypeVar("T")


class AssetToDict(TypedDict):
    assetid: str
    amount: int
    appid: str
    contextid: str


class AssetDict(AssetToDict):
    instanceid: str
    classid: str
    missing: bool


class DescriptionDict(TypedDict, total=False):
    instanceid: Required[str]
    classid: Required[str]
    market_name: str
    currency: int
    name: str
    market_hash_name: str
    name_color: str
    background_color: str  # hex code
    type: str
    descriptions: dict[str, str]
    market_actions: list[dict[str, str]]
    actions: list[dict[str, str]]
    tags: list[dict[str, str]]
    actions: list[dict[str, str]]
    icon_url: str
    icon_url_large: str
    tradable: bool  # 1 vs 0
    marketable: bool  # same as above
    commodity: int  # might be a bool
    fraudwarnings: list[str]


class ItemDict(AssetDict, DescriptionDict):
    """We combine Assets with their matching Description to form items."""


class InventoryDict(TypedDict):
    assets: list[AssetDict]
    descriptions: list[DescriptionDict]
    total_inventory_count: int
    success: int  # Result
    rwgrsn: int  # p. much always -2


class TradeOfferDict(TypedDict):
    tradeofferid: str
    tradeid: str  # only used for receipts (its not the useful one)
    accountid_other: int
    message: str
    trade_offer_state: int  # TradeOfferState
    expiration_time: int  # unix timestamps
    time_created: int
    time_updated: int
    escrow_end_date: int
    items_to_give: list[ItemDict]
    items_to_receive: list[ItemDict]
    is_our_offer: bool
    from_real_time_trade: bool
    confirmation_method: int  # 2 is mobile 1 might be email? not a clue what other values are


class TradeOfferReceiptAssetDict(AssetDict):
    new_assetid: str
    new_contextid: str


class TradeOfferReceiptItemDict(TradeOfferReceiptAssetDict, ItemDict):
    ...


class TradeOfferReceiptDict(TypedDict):
    status: int
    tradeid: str
    time_init: int
    assets_received: NotRequired[list[TradeOfferReceiptAssetDict]]
    assets_given: NotRequired[list[TradeOfferReceiptAssetDict]]
    descriptions: list[DescriptionDict]


class Asset:
    """Base most version of an item. This class should only be received when Steam fails to find a matching item for
    its class and instance IDs.

    .. container:: operations

        .. describe:: x == y

            Checks if two assets are equal.

    Attributes
    -------------
    asset_id
        The assetid of the item.
    amount
        The amount of the same asset there are in the inventory.
    instance_id
        The instanceid of the item.
    class_id
        The classid of the item.
    owner
        The owner of the asset
    """

    __slots__ = ("amount", "class_id", "asset_id", "instance_id", "owner", "_game_cs", "_app_id")
    REPR_ATTRS = ("amount", "class_id", "asset_id", "instance_id", "owner", "game")

    def __init__(self, data: AssetDict, owner: BaseUser):
        self.asset_id = int(data["assetid"])
        self.amount = int(data["amount"])
        self.instance_id = int(data["instanceid"])
        self.class_id = int(data["classid"])
        self.owner = owner
        self._app_id = int(data["appid"])

    def __repr__(self) -> str:
        cls = self.__class__
        resolved = [f"{attr}={getattr(self, attr, None)!r}" for attr in cls.REPR_ATTRS]
        return f"<{cls.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Asset) and self.instance_id == other.instance_id and self.class_id == other.class_id

    def to_dict(self) -> AssetToDict:
        return {
            "assetid": str(self.asset_id),
            "amount": self.amount,
            "appid": str(self.game.id),
            "contextid": str(self.game.context_id),
        }

    @utils.cached_slot_property
    def game(self) -> StatefulGame:
        """The game the item is from."""
        return StatefulGame(self._state, id=self._app_id)

    @property
    def _state(self) -> ConnectionState:
        return self.owner._state


class Item(Asset):
    """Represents an item in an User's inventory.

    .. container:: operations

        .. describe:: x == y

            Checks if two items are equal.

    Attributes
    -------------
    name
        The market_name of the item.
    display_name
        The displayed name of the item. This could be different to :attr:`Item.name` if the item is user re-nameable.
    colour
        The colour of the item.
    descriptions
        The descriptions of the item.
    type
        The type of the item.
    tags
        The tags of the item.
    icon_url
        The icon url of the item. Uses the large (184x184 px) image url.
    fraud_warnings
        The fraud warnings for the item.
    actions
        The actions for the item.
    """

    __slots__ = (
        "name",
        "type",
        "tags",
        "colour",
        "icon_url",
        "display_name",
        "descriptions",
        "fraud_warnings",
        "actions",
        "_is_tradable",
        "_is_marketable",
    )
    REPR_ATTRS = ("name", *Asset.REPR_ATTRS)

    def __init__(self, data: ItemDict, owner: BaseUser):
        super().__init__(data, owner)
        self._from_data(data)

    def _from_data(self, data: ItemDict) -> None:
        self.name = data.get("market_name")
        self.display_name = data.get("name")
        self.colour = int(data["name_color"], 16) if "name_color" in data else None
        self.descriptions = data.get("descriptions")
        self.type = data.get("type")
        self.tags = data.get("tags")
        self.icon_url = (
            f'https://steamcommunity-a.akamaihd.net/economy/image/{data["icon_url_large"]}'
            if "icon_url_large" in data
            else None
        )
        self.fraud_warnings = data.get("fraudwarnings", [])
        self.actions = data.get("actions", [])
        self._is_tradable = bool(data.get("tradable", False))
        self._is_marketable = bool(data.get("marketable", False))

    def is_tradable(self) -> bool:
        """Whether the item is tradable."""
        return self._is_tradable

    def is_marketable(self) -> bool:
        """Whether the item is marketable."""
        return self._is_marketable


if hasattr(types, "GenericAlias"):
    GenericAlias = types.GenericAlias
    kwargs = {}
else:
    GenericAlias = type(
        types.new_class(
            "",
            (Generic[T],),
        )[int]
    )
    kwargs = {"_root": True}


class InventoryGenericAlias(GenericAlias, **kwargs):
    def __repr__(self) -> str:
        return f"{self.__origin__.__module__}.{object.__getattribute__(self, '__alias_name__')}"

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        # this is done cause we need __orig_class__ in __init__
        result = self.__origin__.__new__(self.__origin__, *args, **kwargs)
        try:
            result.__orig_class__ = self
        except AttributeError:
            pass
        result.__init__(*args, **kwargs)
        return result


class BaseInventory(Generic[I]):
    """Base for all inventories."""

    __slots__ = (
        "game",
        "items",
        "owner",
        "_state",
        "__orig_class__",  # undocumented typing internals more shim to make setting __class__ work
    )

    def __init__(self, state: ConnectionState, data: InventoryDict, owner: BaseUser, game: Game):
        self._state = state
        self.owner = owner
        self.game = StatefulGame(state, id=game.id, name=game.name)
        self._update(data)

    def __repr__(self) -> str:
        attrs = ("owner", "game")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{object.__getattribute__(self.__orig_class__, '__alias_name__')} {' '.join(resolved)}>"

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[I]:
        return iter(self.items)

    def __contains__(self, item: Asset) -> bool:
        if not isinstance(item, Asset):
            raise TypeError(
                f"unsupported operand type(s) for 'in': {item.__class__.__qualname__!r} and {self.__class__.__name__!r}"
            )
        return item in self.items

    if not TYPE_CHECKING:

        def __class_getitem__(cls, params: tuple[type[I]]) -> InventoryGenericAlias:
            # this is more stuff that's needed to make the TypeAliases for extension modules work as we need the
            # assigned name

            frame = sys._getframe(1)
            generic_alias = InventoryGenericAlias(cls, params)

            on_line = False
            return_next = False
            for instruction in dis.get_instructions(frame.f_code):
                if return_next and instruction.opname == "STORE_NAME":
                    break
                elif instruction.starts_line == frame.f_lineno:
                    on_line = True
                elif on_line and instruction.opname == "BINARY_SUBSCR":
                    return_next = True
            else:
                return generic_alias

            object.__setattr__(generic_alias, "__alias_name__", instruction.argval)
            return generic_alias

    def _update(self, data: InventoryDict) -> None:
        self.items: Sequence[I] = []
        ItemClass: type[Item] = self.__orig_class__.__args__[0]
        for asset in data.get("assets", ()):
            for item in data["descriptions"]:
                if item["instanceid"] == asset["instanceid"] and item["classid"] == asset["classid"]:
                    item.update(asset)
                    self.items.append(  # type: ignore
                        ItemClass(data=item, owner=self.owner),
                    )
                    break
            else:
                self.items.append(  # type: ignore
                    Asset(data=asset, owner=self.owner),
                )

    async def update(self) -> None:
        """Re-fetches the inventory."""
        data = await self._state.http.get_user_inventory(self.owner.id64, self.game.id, self.game.context_id)
        self._update(data)

    def filter_items(self, *names: str, limit: int | None = None) -> list[I]:
        """A helper function that filters items by name from the inventory.

        Parameters
        ------------
        names
            The names of the items to filter for.
        limit
            The maximum amount of items to return. Checks from the front of the items.

        Raises
        -------
        :exc:`ValueError`
            You passed a limit and multiple item names.

        Returns
        ---------
        The matching items.
        """
        if len(names) > 1 and limit:
            raise ValueError("Cannot pass a limit with multiple items")
        items = [item for item in self if item.name in names]
        return items if limit is None else items[:limit]

    def get_item(self, name: str) -> I | None:
        """A helper function that gets an item or ``None`` if no matching item is found by name from the inventory.

        Parameters
        ----------
        name
            The item to get from the inventory.
        """
        item = self.filter_items(name, limit=1)
        return item[0] if item else None


Inventory: TypeAlias = BaseInventory[Item]  # necessitated by TypeVar not currently supporting defaults
"""Represents a User's inventory.

.. container:: operations

    .. describe:: len(x)

        Returns how many items are in the inventory.

    .. describe:: iter(x)

        Iterates over the inventory's items.

    .. describe:: y in x

        Determines if an item is in the inventory based off of its :attr:`class_id` and :attr:`instance_id`.


Attributes
----------
items
    A list of the inventory's items.
owner
    The owner of the inventory.
game
    The game the inventory the game belongs to.
"""


class TradeOfferReceipt(NamedTuple):
    sent: list[TradeOfferReceiptItem]
    received: list[TradeOfferReceiptItem]


class TradeOfferReceiptItem(Item):
    """An item in a trade receipt.

    Attributes
    ----------
    new_asset_id
        The new_assetid field, this is the asset id of the item in the partners inventory.
    new_context_id
        The new_contextid field.
    """

    __slots__ = (
        "new_asset_id",
        "new_context_id",
    )
    new_asset_id: int
    new_context_id: int

    def _from_data(self, data: TradeOfferReceiptItemDict):
        super()._from_data(data)
        self.new_context_id = int(data["new_contextid"])
        self.new_asset_id = int(data["new_assetid"])


class TradeOffer:
    """Represents a trade offer from/to send to a User.
    This can also be used in :meth:`steam.User.send`.

    Parameters
    ----------
    item_to_send
        The item to send with the trade offer.
    item_to_receive
        The item to receive with the trade offer.
    items_to_send
        The items you are sending to the other user.
    items_to_receive
        The items you are receiving from the other user.
    token
        The the trade token used to send trades to users who aren't on the ClientUser's friend's list.
    message
         The offer message to send with the trade.

    Attributes
    ----------
    partner
        The trade offer partner. This should only ever be a :class:`~steam.SteamID` if the partner's profile is private.
    items_to_send
        A list of items to send to the partner.
    items_to_receive
        A list of items to receive from the partner.
    state
        The offer state of the trade for the possible types see :class:`~steam.TradeOfferState`.
    message
        The message included with the trade offer.
    id
        The trade's offer ID.
    created_at
        The time at which the trade was created.
    updated_at
        The time at which the trade was last updated.
    expires
        The time at which the trade automatically expires.
    escrow
        The time at which the escrow will end. Can be ``None`` if there is no escrow on the trade.

        Warning
        -------
        This isn't likely to be accurate, use :meth:`User.escrow` instead if possible.
    """

    __slots__ = (
        "id",
        "_id",
        "state",
        "escrow",
        "partner",
        "message",
        "token",
        "expires",
        "updated_at",
        "created_at",
        "items_to_send",
        "items_to_receive",
        "_has_been_sent",
        "_state",
        "_is_our_offer",
    )

    def __init__(
        self,
        *,
        message: str | None = None,
        token: str | None = None,
        item_to_send: Items | None = None,
        item_to_receive: Items | None = None,
        items_to_send: list[Items] | None = None,
        items_to_receive: list[Items] | None = None,
    ):
        self.items_to_receive: list[Items] = items_to_receive or []
        self.items_to_send: list[Items] = items_to_send or []
        if item_to_receive:
            self.items_to_receive.append(item_to_receive)
        if item_to_send:
            self.items_to_send.append(item_to_send)
        self.message: str | None = message or None
        self.token: str | None = token
        self.partner: User | SteamID | None = None
        self.updated_at: datetime | None = None
        self.created_at: datetime | None = None
        self.escrow: timedelta | None = None
        self.state = TradeOfferState.Invalid
        self._id: int | None = None
        self._has_been_sent = False

    @classmethod
    def _from_api(
        cls, state: ConnectionState, data: TradeOfferDict, partner: User | SteamID | None = None
    ) -> TradeOffer:
        trade = cls()
        trade._has_been_sent = True
        trade._state = state
        trade.partner = partner or utils.make_id64(data["accountid_other"])  # type: ignore
        trade._update(data)
        return trade

    def _update_from_send(
        self, state: ConnectionState, data: dict[str, Any], partner: User, active: bool = True
    ) -> None:
        self.id = int(data["tradeofferid"])
        self._state = state
        self.partner = partner
        self.state = TradeOfferState.Active if active else TradeOfferState.ConfirmationNeed
        self.created_at = datetime.utcnow()
        self._is_our_offer = True

    def __repr__(self) -> str:
        attrs = ("id", "state", "partner")
        resolved = [f"{attr}={getattr(self, attr, None)!r}" for attr in attrs]
        return f"<TradeOffer {' '.join(resolved)}>"

    def _update(self, data: TradeOfferDict) -> None:
        self.message = data.get("message") or None
        self.id = int(data["tradeofferid"])
        self._id = int(data["tradeid"]) if "tradeid" in data else None
        expires = data.get("expiration_time")
        escrow = data.get("escrow_end_date")
        updated_at = data.get("time_updated")
        created_at = data.get("time_created")
        self.expires = datetime.utcfromtimestamp(expires) if expires else None
        self.escrow = datetime.utcfromtimestamp(escrow) - datetime.utcnow() if escrow else None
        self.updated_at = datetime.utcfromtimestamp(updated_at) if updated_at else None
        self.created_at = datetime.utcfromtimestamp(created_at) if created_at else None
        self.state = TradeOfferState.try_value(data.get("trade_offer_state", 1))
        self.items_to_send = [Item(data=item, owner=self.partner) for item in data.get("items_to_give", [])]
        self.items_to_receive = [Item(data=item, owner=self.partner) for item in data.get("items_to_receive", [])]
        self._is_our_offer = data.get("is_our_offer", False)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, TradeOffer) and self._has_been_sent and other._has_been_sent and self.id == other.id

    async def confirm(self) -> None:
        """Confirms the trade offer.
        This rarely needs to be called as the client handles most of these.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade is not active.
        :exc:`~steam.ConfirmationError`
            No matching confirmation could not be found.
        """
        self._check_active()
        if self.is_gift():
            return  # no point trying to confirm it
        if not await self._state.fetch_and_confirm_confirmation(self.id):
            raise ConfirmationError("No matching confirmation could be found for this trade")
        del self._state._confirmations[self.id]

    async def accept(self) -> None:
        """Accepts the trade offer.

        Note
        ----
        This also calls :meth:`confirm` (if necessary) so you don't have to.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade is either not active, already accepted or not from the ClientUser.
        :exc:`~steam.ConfirmationError`
            No matching confirmation could not be found.
        """
        if self.state == TradeOfferState.Accepted:
            raise ClientException("This trade has already been accepted")
        if self.is_our_offer():
            raise ClientException("You cannot accept an offer the ClientUser has made")
        self._check_active()
        resp = await self._state.http.accept_user_trade(self.partner.id64, self.id)
        if resp.get("needs_mobile_confirmation", False):
            for tries in range(5):
                try:
                    await self.confirm()
                except ConfirmationError:
                    break
                except ClientException:
                    await asyncio.sleep(tries * 2)

    async def decline(self) -> None:
        """Declines the trade offer.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade is either not active, already declined or not from the ClientUser.
        """
        if self.state == TradeOfferState.Declined:
            raise ClientException("This trade has already been declined")
        if self.is_our_offer():
            raise ClientException("You cannot decline an offer the ClientUser has made")
        self._check_active()
        await self._state.http.decline_user_trade(self.id)

    async def cancel(self) -> None:
        """Cancels the trade offer.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade is either not active or already cancelled.
        """
        if self.state == TradeOfferState.Canceled:
            raise ClientException("This trade has already been cancelled")
        self._check_active()
        await self._state.http.cancel_user_trade(self.id)

    async def receipt(self) -> TradeOfferReceipt:
        """Get the receipt for a trade offer and the updated asset ids for the trade.

        Returns
        -------
        A trade receipt.

        .. source:: steam.TradeOfferReceipt
        """
        if self._id is None:
            raise ValueError("Cannot fetch the receipt for a trade not accepted")

        resp = await self._state.http.get_trade_receipt(self._id)
        data = resp["response"]
        trade = data["trades"][0]
        descriptions = data["descriptions"]

        received: list[TradeOfferReceiptItem] = []
        for asset in trade.get("assets_received", ()):
            for item in descriptions:
                if item["instanceid"] == asset["instanceid"] and item["classid"] == asset["classid"]:
                    item.update(asset)
                    received.append(TradeOfferReceiptItem(data=item, owner=self.partner))  # type: ignore

        sent: list[TradeOfferReceiptItem] = []
        for asset in trade.get("assets_given", ()):
            for item in descriptions:
                if item["instanceid"] == asset["instanceid"] and item["classid"] == asset["classid"]:
                    item.update(asset)
                    sent.append(TradeOfferReceiptItem(data=item, owner=self._state.http.user))  # type: ignore

        return TradeOfferReceipt(sent=sent, received=received)

    async def counter(self, trade: TradeOffer) -> None:
        """Counter a trade offer from an :class:`User`.

        Parameters
        -----------
        trade
            The trade offer to counter with.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade from the ClientUser or it isn't active.
        """
        self._check_active()
        if self.is_our_offer():
            raise ClientException("You cannot counter an offer the ClientUser has made")

        to_send = [item.to_dict() for item in trade.items_to_send]
        to_receive = [item.to_dict() for item in trade.items_to_receive]
        resp = await self._state.http.send_trade_offer(
            self.partner, to_send, to_receive, trade.token, trade.message or "", trade_id=self.id
        )
        if resp.get("needs_mobile_confirmation", False):
            await self._state.fetch_and_confirm_confirmation(int(resp["tradeofferid"]))

    def is_gift(self) -> bool:
        """Helper method that checks if an offer is a gift to the :class:`~steam.ClientUser`"""
        return bool(self.items_to_receive and not self.items_to_send)

    def is_our_offer(self) -> bool:
        """Whether the offer was created by the :class:`~steam.ClientUser`."""
        return self._is_our_offer

    def _check_active(self) -> None:
        if self.state not in (TradeOfferState.Active, TradeOfferState.ConfirmationNeed) or not self._has_been_sent:
            raise ClientException("This trade is not active")



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import math
import re
from collections import deque
from collections.abc import Callable, Coroutine
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from bs4 import BeautifulSoup
from typing_extensions import ClassVar, TypeAlias

from . import utils
from .comment import Comment

if TYPE_CHECKING:
    from .abc import Channel, Commentable, Message
    from .channel import ClanChannel, ClanMessage, DMChannel, GroupChannel, GroupMessage, UserMessage
    from .clan import Clan
    from .event import Announcement, Event
    from .game import StatefulGame
    from .group import Group
    from .state import ConnectionState
    from .trade import DescriptionDict, TradeOffer

T = TypeVar("T")
TT = TypeVar("TT")
ChannelT = TypeVar("ChannelT", bound="Channel[Any]", covariant=True)
CommentableT = TypeVar("CommentableT", bound="Commentable")
M = TypeVar("M", bound="Message", covariant=True)

MaybeCoro: TypeAlias = "Callable[[T], bool | Coroutine[Any, Any, bool]]"
UNIX_EPOCH = datetime.utcfromtimestamp(0)


class AsyncIterator(Generic[T]):  # TODO re-work to be fetch in chunks in V1
    """A class from which async iterators (see :pep:`525`) can ben easily derived.

    .. container:: operations

        .. describe:: async for y in x

            Iterates over the contents of the async iterator.

    Attributes
    ----------
    before
        When to find objects before.
    after
        When to find objects after.
    limit
        The maximum size of the :attr:`queue`.
    queue
        The queue containing the elements of the iterator.
    """

    def __init__(
        self,
        state: ConnectionState,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ):
        self._state = state
        self.before = before or datetime.utcnow()
        self.after = after or UNIX_EPOCH
        self._is_filled = False
        self.queue: deque[T] = deque()
        self.limit = limit

    def _append(self, element: T) -> bool:
        if self.limit is None:
            self.queue.append(element)
            return True
        if len(self.queue) < self.limit:
            self.queue.append(element)
            return True
        if len(self.queue) == self.limit:
            self.queue.append(element)

        return False

    async def _fill_queue_users(
        self,
        id64s: set[Any],  # should be set[int] but # type: ignore stuff forces this
        attributes: tuple[str, ...] = ("author",),
    ) -> None:
        for user in await self._state.fetch_users(list(id64s)):
            for element in self.queue:
                for attribute in attributes:
                    if getattr(element, attribute, None) == user:
                        setattr(element, attribute, user)

    async def get(self, **attrs: Any) -> T | None:
        """A helper function which is similar to :func:`~steam.utils.get` except it runs over the async iterator.

        This is roughly equivalent to:

        .. code-block:: python3


            elements = await async_iterator.flatten()
            element = steam.utils.get(elements, name="Item")

        Example
        -------
        Getting the last comment from a user named 'Dave' or None:

        .. code-block:: python3

            comment = await user.comments().get(author__name="Dave")

        Parameters
        ----------
        attrs
            Keyword arguments that denote attributes to match.

        Returns
        -------
        The first element from the ``iterable`` which matches all the traits passed in ``attrs`` or ``None`` if no
        matching element was found.
        """

        def predicate(elem: T) -> bool:
            for attr, val in attrs.items():
                nested = attr.split("__")
                obj = elem
                for attribute in nested:
                    obj = getattr(obj, attribute)

                if obj != val:
                    return False
            return True

        return await self.find(predicate)

    async def find(self, predicate: MaybeCoro[T]) -> T | None:
        """A helper function which is similar to :func:`~steam.utils.find` except it runs over the async iterator.
        However unlike :func:`~steam.utils.find`, the predicate provided can be a |coroutine_link|_.

        This is roughly equivalent to:

        .. code-block:: python3

            elements = await async_iterator.flatten()
            element = steam.utils.find(elements, lambda e: e.name == "Item")

        Example
        -------
        Getting the last trade with a message or None:

        .. code-block:: python3

            def predicate(trade: steam.TradeOffer) -> bool:
                return trade.message is not None


            trade = await client.trade_history().find(predicate)

        Parameters
        ----------
        predicate
            A callable/coroutine that returns a boolean.

        Returns
        -------
        The first element from the iterator for which the ``predicate`` returns ``True`` or ``None`` if no matching
        element was found.
        """
        async for elem in self:
            ret = await utils.maybe_coroutine(predicate, elem)
            if ret:
                return elem

    async def flatten(self) -> list[T]:
        """A helper function that iterates over the :class:`AsyncIterator` returning a list of all the elements in the
        iterator.

        This is equivalent to:

        .. code-block:: python3

            elements = [element async for element in async_iterator]
        """
        return [element async for element in self]

    def filter(self, predicate: Callable[[T], bool]) -> FilteredIterator[T]:
        """Filter members of the async iterator according to a predicate. This function acts similarly to :func:`filter`.

        Examples
        --------
        .. code-block:: python3

            for dave in async_iterator.filter(lambda x: x.name == "Dave"):
                ...  # the element now has to have a name of Dave.

        Parameters
        ----------
        predicate
            The predicate to filter elements through.
        """
        return FilteredIterator(predicate, self)

    def map(self, func: Callable[[TT], Any]) -> MappedIterator[T, TT]:
        """Map the elements of the async iterator through a function. This function acts similarly to :func:`map`.

        Examples
        --------
        .. code-block:: python3

            for name in async_iterator.map(lambda x: x.name):
                ...  # name is now the iterators element's name.

        Parameters
        ----------
        func
            The function to map the elements through.
        """
        return MappedIterator(func, self)

    def __aiter__(self) -> AsyncIterator[T]:
        return self

    def __anext__(self) -> Coroutine[None, None, T]:
        return self.next()

    async def next(self) -> T:
        """Advances the iterator by one, if possible.

        Raises
        ------
        :exc:`StopAsyncIteration`
            There are no more elements in the iterator.
        """
        if not self.queue:
            if self._is_filled:
                raise StopAsyncIteration
            await self.fill()
            self._is_filled = True
        if not self.queue:  # yikes
            raise StopAsyncIteration
        return self.queue.pop()

    async def fill(self) -> None:
        raise NotImplementedError


class FilteredIterator(AsyncIterator[T]):
    def __init__(self, predicate: MaybeCoro[T], async_iterator: AsyncIterator[T]):
        self.predicate = predicate
        self.iterator = async_iterator

    async def next(self) -> T:
        while True:
            item = await self.iterator.next()
            if await utils.maybe_coroutine(self.predicate, item):
                return item


class MappedIterator(AsyncIterator[TT], Generic[T, TT]):
    def __init__(self, map_func: Callable[[Any], TT | Coroutine[Any, Any, TT]], async_iterator: AsyncIterator[T]):
        self.map_func = map_func
        self.iterator = async_iterator

    async def next(self) -> TT:
        item = await self.iterator.next()
        return await utils.maybe_coroutine(self.map_func, item)


class CommentsIterator(AsyncIterator[Comment[CommentableT]]):
    def __init__(
        self,
        owner: CommentableT,
        oldest_first: bool,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        self.owner = owner
        self.oldest_first = oldest_first
        super().__init__(state, limit, before, after)

    async def fill(self) -> None:
        comments = await self._state.fetch_comments(self.owner, self.limit, self.after, self.oldest_first)
        author_id64s = set()
        for comment in comments:
            comment = Comment(
                self._state,
                id=comment.id,
                content=comment.content,
                created_at=datetime.utcfromtimestamp(comment.timestamp),
                author=comment.author_id64,  # type: ignore
                owner=self.owner,
            )
            if not self.after < comment.created_at < self.before:
                continue
            if not self._append(comment):
                break
            author_id64s.add(comment.author)

        await self._fill_queue_users(author_id64s)


class TradesIterator(AsyncIterator["TradeOffer"]):
    def __init__(
        self,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        super().__init__(state, limit, before, after)

    async def fill(self) -> None:
        from .trade import TradeOffer

        resp = await self._state.http.get_trade_history(100, None)
        resp = resp["response"]
        total = resp.get("total_trades", 0)
        if not total:
            return

        descriptions = resp.get("descriptions", [])
        after_timestamp = self.after.timestamp()
        before_timestamp = self.before.timestamp()

        class Stop(Exception):
            ...

        async def process_trade(data: dict[str, Any], descriptions: list[DescriptionDict]) -> None:
            if not after_timestamp < data["time_init"] < before_timestamp:
                return
            for item in descriptions:
                for asset in data.get("assets_received", []):
                    if item["classid"] == asset["classid"] and item["instanceid"] == asset["instanceid"]:
                        asset.update(item)
                for asset in data.get("assets_given", []):
                    if item["classid"] == asset["classid"] and item["instanceid"] == asset["instanceid"]:
                        asset.update(item)

            # patch in the attributes cause steam is cool
            data["tradeofferid"] = data["tradeid"]
            data["accountid_other"] = data["steamid_other"]
            data["trade_offer_state"] = data["status"]
            data["items_to_give"] = data.get("assets_given", [])
            data["items_to_receive"] = data.get("assets_received", [])

            trade = TradeOffer._from_api(state=self._state, data=data)

            if not self._append(trade):
                raise Stop
            partner_id64s.add(trade.partner)

        partner_id64s = set()

        try:
            for trade in resp.get("trades", []):
                await process_trade(trade, descriptions)

            previous_time = trade["time_init"]
            if total < 100:
                for page in range(200, math.ceil((total + 100) / 100) * 100, 100):
                    resp = await self._state.http.get_trade_history(page, previous_time)
                    resp = resp["response"]

                    for trade in resp.get("trades", []):
                        previous_time = trade["time_init"]
                        await process_trade(trade, descriptions)
        except Stop:
            pass

        for user in await self._state.fetch_users(list(partner_id64s)):
            for element in self.queue:
                if element.partner == user:
                    element.partner = user
                    for item in element.items_to_receive:
                        item.owner = user
                for item in element.items_to_send:
                    item.owner = self._state.http.user


class ChannelHistoryIterator(AsyncIterator[M], Generic[M, ChannelT]):
    def __init__(
        self,
        channel: ChannelT,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        super().__init__(state, limit, before, after)
        self.before = before or UNIX_EPOCH
        self.channel = channel


class DMChannelHistoryIterator(ChannelHistoryIterator["UserMessage", "DMChannel"]):
    __slots__ = ("participant",)

    def __init__(
        self,
        channel: DMChannel,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        super().__init__(channel, state, limit, before, after)
        self.participant = channel.participant

    async def fill(self) -> None:
        from .message import Message, UserMessage

        after_timestamp = int(self.after.timestamp())
        before_timestamp = int(self.before.timestamp())

        last_message_timestamp = before_timestamp

        while True:
            resp = await self._state.fetch_user_history(
                self.participant.id64, start=after_timestamp, last=last_message_timestamp
            )
            if not resp.messages:
                return

            for message in resp.messages:
                new_message = UserMessage.__new__(UserMessage)
                Message.__init__(new_message, channel=self.channel, proto=message)
                new_message.author = (  # type: ignore
                    self.participant if message.accountid == self.participant.id else self._state.client.user
                )
                new_message.created_at = datetime.utcfromtimestamp(message.timestamp)
                if not self._append(new_message):
                    return

            last_message_timestamp = int(message.timestamp)

            if not resp.more_available:
                return


GroupMessages = TypeVar("GroupMessages", bound="ClanMessage | GroupMessage")
GroupChannels = TypeVar("GroupChannels", bound="ClanChannel | GroupChannel")


class GroupChannelHistoryIterator(ChannelHistoryIterator[GroupMessages, GroupChannels]):
    __slots__ = ("group",)

    def __init__(
        self,
        channel: ClanChannel | GroupChannel,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        super().__init__(channel, state, limit, before, after)
        self.group: Group | Clan = channel.group or channel.clan  # type: ignore

    async def fill(self) -> None:
        from .message import ClanMessage, GroupMessage, Message

        after_timestamp = int(self.after.timestamp())
        before_timestamp = int(self.before.timestamp())
        last_message_timestamp = before_timestamp
        group_id = getattr(self.group, "chat_id", None) or self.group.id
        author_id64s = set()

        while True:
            resp = await self._state.fetch_group_history(
                group_id, self.channel.id, start=after_timestamp, last=last_message_timestamp
            )
            if not resp.messages:
                return

            for message in resp.messages:
                new_message = (
                    GroupMessage.__new__(GroupMessage) if self.channel.group else ClanMessage.__new__(ClanMessage)
                )
                Message.__init__(new_message, channel=self.channel, proto=message)
                new_message.author = utils.make_id64(message.sender)  # type: ignore
                author_id64s.add(new_message.author)
                new_message.created_at = datetime.utcfromtimestamp(message.server_timestamp)
                if not self._append(new_message):
                    break

            last_message_timestamp = int(message.server_timestamp)

            if not resp.more_available:
                break

        await self._fill_queue_users(author_id64s)


class _EventIterator(AsyncIterator[T]):
    ID_PARSE_REGEX: ClassVar[re.Pattern[str]]

    def __init__(
        self, clan: Clan, state: ConnectionState, limit: int | None, before: datetime | None, after: datetime | None
    ):
        super().__init__(state, limit, before, after)
        self.clan = clan

    async def fill(self) -> None:
        cls = self.__class__
        rss = await self._state.http.get_clan_rss(
            self.clan.id64
        )  # TODO make this use the calendar? does that work for announcements

        soup = BeautifulSoup(rss, "html.parser")

        ids = []
        for url in soup.find_all("guid"):
            match = cls.ID_PARSE_REGEX.findall(url.text)
            if match:
                ids.append(int(match[0]))

        if not ids:
            return

        events = await self.get_events(ids)
        to_fetch_id64s = set()
        from . import event

        event_cls: type[Announcement | Event] = getattr(
            event, cls.__orig_bases__[0].__args__[0].__forward_arg__  # type: ignore
        )

        for event_ in events["events"]:
            event = event_cls(self._state, self.clan, event_)
            to_fetch_id64s.add(event.author)
            to_fetch_id64s.add(event.last_edited_by)
            if not self._append(event):
                break

        await self._fill_queue_users(to_fetch_id64s, ("author", "last_edited_by", "approved_by"))

    async def get_events(self, ids: list[int]) -> dict[str, Any]:
        raise NotImplementedError


class EventIterator(_EventIterator["Event"]):
    ID_PARSE_REGEX = re.compile(r"events/+(\d+)")

    async def get_events(self, ids: list[int]) -> dict[str, Any]:
        return await self._state.http.get_clan_events(self.clan.id, ids)


class AnnouncementsIterator(_EventIterator["Announcement"]):
    ID_PARSE_REGEX = re.compile(r"announcements/detail/(\d+)")

    async def get_events(self, ids: list[int]) -> dict[str, Any]:
        announcements = await asyncio.gather(*(self._state.http.get_clan_announcement(self.clan.id, id) for id in ids))
        events = []
        for announcement in announcements:
            events += announcement["events"]
        return {"events": events}



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import logging
import re
import sys
import traceback
from collections.abc import Callable
from datetime import timedelta
from typing import TYPE_CHECKING, Any, TypeVar

from typing_extensions import Final, Literal, ParamSpec, TypedDict
from yarl import URL as _URL

from . import utils
from .enums import IntEnum
from .protobufs import EMsg, chat

if TYPE_CHECKING:
    from .gateway import Msgs


__all__ = (
    "PriceOverview",
    "Ban",
)

F = TypeVar("F", bound="Callable[..., Any]")
R = TypeVar("R", bound="Registerable")
P = ParamSpec("P")
TASK_HAS_NAME = sys.version_info >= (3, 8)


def api_route(path: str, version: int = 1) -> _URL:
    return URL.API / f"{path}/v{version}"


class URL:
    API: Final = _URL("https://api.steampowered.com")
    COMMUNITY: Final = _URL("https://steamcommunity.com")
    STORE: Final = _URL("https://store.steampowered.com")


class _ReturnTrue:
    __slots__ = ()

    def __call__(self, *_: Any, **__: Any) -> Literal[True]:
        return True

    def __repr__(self) -> str:
        return "<return_true>"


return_true = _ReturnTrue()


class Registerable:
    __slots__ = ("loop", "parsers_name")

    def __new__(cls: type[R], *args: Any, **kwargs: Any) -> R:
        self = super().__new__(cls)
        self.loop = asyncio.get_event_loop()
        cls.parsers_name = tuple(cls.__annotations__)[0]
        bases = tuple(reversed(cls.__mro__[:-2]))  # skip Registerable and object
        for idx, base in enumerate(bases):
            parsers_name = tuple(base.__annotations__)[0]
            for name, attr in base.__dict__.items():
                if not hasattr(attr, "msg"):
                    continue
                try:
                    parsers = getattr(self, parsers_name)
                except AttributeError:
                    parsers = {}
                    setattr(self, parsers_name, parsers)
                msg_parser = getattr(self, name)
                if idx != 0 and isinstance(attr.msg, EMsg):
                    parsers = getattr(self, tuple(bases[0].__annotations__)[0])
                parsers[attr.msg] = msg_parser

        return self

    @staticmethod
    def _run_parser_callback(task: asyncio.Task[None]) -> None:
        exception = task.exception()
        if exception:
            traceback.print_exception(exception.__class__, exception, exception.__traceback__)

    def run_parser(self, emsg: IntEnum, msg: Msgs) -> None:
        try:
            event_parser = getattr(self, self.parsers_name)[emsg]
        except KeyError:
            log = logging.getLogger(self.__class__.__module__)
            if log.isEnabledFor(logging.DEBUG):
                try:
                    log.debug(f"Ignoring event {msg!r}")
                except Exception:
                    log.debug(f"Ignoring event {msg.msg}")
        else:
            coro = utils.maybe_coroutine(event_parser, msg)

            (
                self.loop.create_task(coro, name=f"task_{event_parser.__name__}")
                if TASK_HAS_NAME
                else self.loop.create_task(coro)
            ).add_done_callback(self._run_parser_callback)


EventParser = None


def register(msg: IntEnum) -> Callable[[F], F]:  # this afaict is not type able currently without HKT
    def wrapper(callback: F) -> F:
        callback.msg = msg
        return callback

    return wrapper


PRICE_RE = re.compile(r"(^\D*(?P<price>[\d,.]*)\D*$)")


class PriceOverviewDict(TypedDict):
    success: bool
    lowest_price: str
    median_price: str
    volume: str


class PriceOverview:
    """Represents the data received from https://steamcommunity.com/market/priceoverview.

    Attributes
    -------------
    currency
        The currency identifier for the item e.g. "$" or "£".
    volume
        The amount of items are currently on the market.
    lowest_price
        The lowest price observed by the market.
    median_price
        The median price observed by the market.
    """

    __slots__ = ("currency", "volume", "lowest_price", "median_price")

    lowest_price: float | str
    median_price: float | str

    def __init__(self, data: PriceOverviewDict):
        lowest_price = PRICE_RE.search(data["lowest_price"])["price"]
        median_price = PRICE_RE.search(data["median_price"])["price"]

        try:
            self.lowest_price = float(lowest_price.replace(",", "."))
            self.median_price = float(median_price.replace(",", "."))
        except ValueError:
            self.lowest_price = lowest_price
            self.median_price = median_price

        self.volume: int = int(data["volume"].replace(",", ""))
        self.currency: str = data["lowest_price"].replace(lowest_price, "").strip()

    def __repr__(self) -> str:
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in self.__slots__]
        return f"<PriceOverview {' '.join(resolved)}>"


class Ban:
    """Represents a Steam ban.

    Attributes
    ----------
    since_last_ban
        How many days since the user was last banned
    number_of_game_bans
        The number of game bans the user has.
    """

    __slots__ = (
        "since_last_ban",
        "number_of_game_bans",
        "_vac_banned",
        "_community_banned",
        "_market_banned",
    )

    def __init__(self, data: dict[str, Any]):
        self._vac_banned: bool = data["VACBanned"]
        self._community_banned: bool = data["CommunityBanned"]
        self._market_banned: bool = data["EconomyBan"]
        self.since_last_ban = timedelta(days=data["DaysSinceLastBan"])
        self.number_of_game_bans: int = data["NumberOfGameBans"]

    def __repr__(self) -> str:
        attrs = [
            ("number_of_game_bans", self.number_of_game_bans),
            ("is_vac_banned()", self.is_vac_banned()),
            ("is_community_banned()", self.is_community_banned()),
            ("is_market_banned()", self.is_market_banned()),
        ]
        resolved = [f"{method}={value!r}" for method, value in attrs]
        return f"<Ban {' '.join(resolved)}>"

    def is_banned(self) -> bool:
        """Species if the user is banned from any part of Steam."""
        return any((self.is_vac_banned(), self.is_community_banned(), self.is_market_banned()))

    def is_vac_banned(self) -> bool:
        """Whether or not the user is VAC banned."""
        return self._vac_banned

    def is_community_banned(self) -> bool:
        """Whether or not the user is community banned."""
        return self._community_banned

    def is_market_banned(self) -> bool:
        """Whether or not the user is market banned."""
        return self._market_banned



"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator
from contextlib import asynccontextmanager
from datetime import datetime
from typing import TYPE_CHECKING, Any

from typing_extensions import Self, TypeAlias

from .abc import Channel, M_co, Message, SteamID
from .iterators import DMChannelHistoryIterator, GroupChannelHistoryIterator
from .protobufs.chat import ChatRoomState, IncomingChatMessageNotification, State

if TYPE_CHECKING:
    from .clan import Clan
    from .group import Group
    from .image import Image
    from .message import ClanMessage, GroupMessage, UserMessage
    from .state import ConnectionState
    from .user import User

__all__ = (
    "DMChannel",
    "GroupChannel",
    "ClanChannel",
)


class DMChannel(Channel["UserMessage"]):  # TODO cache these to add last_message.
    """Represents the channel a DM is sent in.

    Attributes
    ----------
    participant
        The recipient of any messages sent.
    """

    __slots__ = ("participant",)
    clan: None
    group: None

    def __init__(self, state: ConnectionState, participant: User):
        super().__init__(state)
        self.participant = participant
        self.clan = None
        self.group = None

    def __repr__(self) -> str:
        return f"<DMChannel participant={self.participant!r}>"

    def _message_func(self, content: str) -> Coroutine[Any, Any, UserMessage]:
        return self.participant._message_func(content)

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self.participant._image_func(image)

    @asynccontextmanager
    async def typing(self) -> Generator[None, None, None]:
        """Send a typing indicator continuously to the channel while in the context manager.

        Note
        ----
        This only works in DMs.

        Usage:

        .. code-block:: python3

            async with ctx.channel.typing():
                ...  # do your expensive operations
        """

        async def inner() -> None:
            while True:
                await asyncio.sleep(10)
                await self.trigger_typing()

        await self.trigger_typing()
        task = self._state.loop.create_task(inner())
        yield
        task.cancel()

    async def trigger_typing(self) -> None:
        """Send a typing indicator to the channel once.

        Note
        ----
        This only works in DMs.
        """
        await self._state.send_user_typing(self.participant.id64)

    def history(
        self,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> DMChannelHistoryIterator:
        return DMChannelHistoryIterator(state=self._state, channel=self, limit=limit, before=before, after=after)


GroupChannelProtos: TypeAlias = "IncomingChatMessageNotification | State | ChatRoomState"


class _GroupChannel(Channel[M_co]):
    __slots__ = ("id", "name", "joined_at", "position", "last_message")

    def __init__(self, state: ConnectionState, proto: GroupChannelProtos):
        super().__init__(state)
        self.id = int(proto.chat_id)
        self.name: str | None = None
        self.joined_at: datetime | None = None
        self.position: int | None = None
        self.last_message: M_co | None = None
        self._update(proto)

    def _update(self, proto: GroupChannelProtos):
        if hasattr(proto, "chat_name"):
            first, _, second = proto.chat_name.partition(" | ")
            name = second or first
        else:
            name = self.name
        self.name = name
        self.joined_at = (
            datetime.utcfromtimestamp(int(proto.time_joined)) if hasattr(proto, "time_joined") else self.joined_at
        )
        self.position = getattr(proto, "sort_order", None) or self.position
        from .message import ClanMessage, GroupMessage

        if isinstance(proto, IncomingChatMessageNotification):
            steam_id = SteamID(proto.steamid_sender)
            last_message = (ClanMessage if isinstance(self, ClanChannel) else GroupMessage)(
                proto, self, self._state.get_user(steam_id.id64) or steam_id  # type: ignore
            )
        elif isinstance(proto, State):
            steam_id = SteamID(proto.accountid_last_message)
            cls = ClanMessage if isinstance(self, ClanChannel) else GroupMessage
            last_message = cls.__new__(cls)
            last_message.author = self._state.get_user(steam_id.id64) or steam_id
            last_message.channel = self  # type: ignore
            last_message.created_at = datetime.utcfromtimestamp(proto.time_last_message)
            proto.message = proto.last_message
            proto.ordinal = NotImplemented
            Message.__init__(last_message, self, proto)
        else:
            last_message = self.last_message
        self.last_message = last_message  # type: ignore

    def __repr__(self) -> str:
        cls = self.__class__
        attrs = ("name", "id", "group" if self.group is not None else "clan", "position")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{cls.__name__} {' '.join(resolved)}>"

    def history(
        self,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> GroupChannelHistoryIterator[M_co, Self]:
        return GroupChannelHistoryIterator(state=self._state, channel=self, limit=limit, before=before, after=after)


class GroupChannel(_GroupChannel["GroupMessage"]):
    """Represents a group channel.

    Attributes
    ----------
    id
        The ID of the channel.
    name
        The name of the channel, this could be the same as the :attr:`~steam.Group.name` if it's the main channel.
    group
        The group to which messages are sent.
    joined_at
        The time the client joined the chat.
    position
        The position of the channel in the channel list.
    last_message
        The last message sent in the channel.
    """

    clan: None

    def __init__(self, state: ConnectionState, group: Group, proto: GroupChannelProtos):
        super().__init__(state, proto)
        self.group: Group = group

    def _message_func(self, content: str) -> Coroutine[Any, Any, GroupMessage]:
        return self._state.send_group_message(self.id, self.group.id, content)  # type: ignore

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self._state.http.send_group_image(self.id, self.group.id, image)


class ClanChannel(_GroupChannel["ClanMessage"]):  # they're basically the same thing
    """Represents a group channel.

    Attributes
    ----------
    id
        The ID of the channel.
    name
        The name of the channel, this could be the same as the :attr:`~steam.Clan.name` if it's the main channel.
    clan
        The clan to which messages are sent.
    joined_at
        The time the client joined the chat.
    position
        The position of the channel in the channel list.
    last_message
        The last message sent in the channel.
    """

    group: None

    def __init__(self, state: ConnectionState, clan: Clan, proto: GroupChannelProtos):
        super().__init__(state, proto)
        self.clan: Clan = clan

    def _message_func(self, content: str) -> Coroutine[Any, Any, ClanMessage]:
        return self._state.send_group_message(self.id, self.clan.chat_id, content)  # type: ignore

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self._state.http.send_group_image(self.id, self.clan.chat_id, image)  # type: ignore



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_appinfo.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class CMsgClientAppInfoUpdate(betterproto.Message):
    last_changenumber: int = betterproto.uint32_field(1)
    send_changelist: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientAppInfoChanges(betterproto.Message):
    current_change_number: int = betterproto.uint32_field(1)
    force_full_update: bool = betterproto.bool_field(2)
    app_ids: List[int] = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientAppInfoRequest(betterproto.Message):
    apps: List["CMsgClientAppInfoRequestApp"] = betterproto.message_field(1)
    supports_batches: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientAppInfoRequestApp(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    section_flags: int = betterproto.uint32_field(2)
    section_crc: List[int] = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientAppInfoResponse(betterproto.Message):
    apps: List["CMsgClientAppInfoResponseApp"] = betterproto.message_field(1)
    apps_unknown: List[int] = betterproto.uint32_field(2)
    apps_pending: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientAppInfoResponseApp(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    change_number: int = betterproto.uint32_field(2)
    sections: List["CMsgClientAppInfoResponseAppSection"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientAppInfoResponseAppSection(betterproto.Message):
    section_id: int = betterproto.uint32_field(1)
    section_kv: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPackageInfoRequest(betterproto.Message):
    package_ids: List[int] = betterproto.uint32_field(1)
    meta_data_only: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPackageInfoResponse(betterproto.Message):
    packages: List["CMsgClientPackageInfoResponsePackage"] = betterproto.message_field(1)
    packages_unknown: List[int] = betterproto.uint32_field(2)
    packages_pending: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientPackageInfoResponsePackage(betterproto.Message):
    package_id: int = betterproto.uint32_field(1)
    change_number: int = betterproto.uint32_field(2)
    sha: bytes = betterproto.bytes_field(3)
    buffer: bytes = betterproto.bytes_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientPicsChangesSinceRequest(betterproto.Message):
    since_change_number: int = betterproto.uint32_field(1)
    send_app_info_changes: bool = betterproto.bool_field(2)
    send_package_info_changes: bool = betterproto.bool_field(3)
    num_app_info_cached: int = betterproto.uint32_field(4)
    num_package_info_cached: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientPicsChangesSinceResponse(betterproto.Message):
    current_change_number: int = betterproto.uint32_field(1)
    since_change_number: int = betterproto.uint32_field(2)
    force_full_update: bool = betterproto.bool_field(3)
    package_changes: List["CMsgClientPicsChangesSinceResponsePackageChange"] = betterproto.message_field(4)
    app_changes: List["CMsgClientPicsChangesSinceResponseAppChange"] = betterproto.message_field(5)
    force_full_app_update: bool = betterproto.bool_field(6)
    force_full_package_update: bool = betterproto.bool_field(7)


@dataclass(eq=False, repr=False)
class CMsgClientPicsChangesSinceResponsePackageChange(betterproto.Message):
    packageid: int = betterproto.uint32_field(1)
    change_number: int = betterproto.uint32_field(2)
    needs_token: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientPicsChangesSinceResponseAppChange(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    change_number: int = betterproto.uint32_field(2)
    needs_token: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientPicsProductInfoRequest(betterproto.Message):
    packages: List["CMsgClientPicsProductInfoRequestPackageInfo"] = betterproto.message_field(1)
    apps: List["CMsgClientPicsProductInfoRequestAppInfo"] = betterproto.message_field(2)
    meta_data_only: bool = betterproto.bool_field(3)
    num_prev_failed: int = betterproto.uint32_field(4)
    supports_package_tokens: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientPicsProductInfoRequestAppInfo(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    access_token: int = betterproto.uint64_field(2)
    only_public_obsolete: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientPicsProductInfoRequestPackageInfo(betterproto.Message):
    packageid: int = betterproto.uint32_field(1)
    access_token: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPicsProductInfoResponse(betterproto.Message):
    apps: List["CMsgClientPicsProductInfoResponseAppInfo"] = betterproto.message_field(1)
    unknown_appids: List[int] = betterproto.uint32_field(2)
    packages: List["CMsgClientPicsProductInfoResponsePackageInfo"] = betterproto.message_field(3)
    unknown_packageids: List[int] = betterproto.uint32_field(4)
    meta_data_only: bool = betterproto.bool_field(5)
    response_pending: bool = betterproto.bool_field(6)
    http_min_size: int = betterproto.uint32_field(7)
    http_host: str = betterproto.string_field(8)


@dataclass(eq=False, repr=False)
class CMsgClientPicsProductInfoResponseAppInfo(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    change_number: int = betterproto.uint32_field(2)
    missing_token: bool = betterproto.bool_field(3)
    sha: bytes = betterproto.bytes_field(4)
    buffer: bytes = betterproto.bytes_field(5)
    only_public: bool = betterproto.bool_field(6)
    size: int = betterproto.uint32_field(7)


@dataclass(eq=False, repr=False)
class CMsgClientPicsProductInfoResponsePackageInfo(betterproto.Message):
    packageid: int = betterproto.uint32_field(1)
    change_number: int = betterproto.uint32_field(2)
    missing_token: bool = betterproto.bool_field(3)
    sha: bytes = betterproto.bytes_field(4)
    buffer: bytes = betterproto.bytes_field(5)
    size: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class CMsgClientPicsAccessTokenRequest(betterproto.Message):
    packageids: List[int] = betterproto.uint32_field(1)
    appids: List[int] = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPicsAccessTokenResponse(betterproto.Message):
    package_access_tokens: List["CMsgClientPicsAccessTokenResponsePackageToken"] = betterproto.message_field(1)
    package_denied_tokens: List[int] = betterproto.uint32_field(2)
    app_access_tokens: List["CMsgClientPicsAccessTokenResponseAppToken"] = betterproto.message_field(3)
    app_denied_tokens: List[int] = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientPicsAccessTokenResponsePackageToken(betterproto.Message):
    packageid: int = betterproto.uint32_field(1)
    access_token: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPicsAccessTokenResponseAppToken(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    access_token: int = betterproto.uint64_field(2)



from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable, Coroutine
from typing import TYPE_CHECKING, Any, ClassVar

from ... import utils
from ...abc import BaseUser
from ...enums import IntEnum
from ...models import register
from ...protobufs import EMsg, GCMsg, GCMsgProto, MsgProto
from ...state import ConnectionState
from ...trade import Inventory

if TYPE_CHECKING:
    from steam.protobufs.client_server_2 import CMsgGcClient

    from ...game import Game
    from .client import Client

log = logging.getLogger(__name__)


class GCState(ConnectionState):
    Language: ClassVar[IntEnum]
    gc_parsers: dict[IntEnum, Callable[..., Any]]
    client: Client

    def __init__(self, client: Client, **kwargs: Any):
        super().__init__(client, **kwargs)
        self._gc_connected = asyncio.Event()
        self._gc_ready = asyncio.Event()
        self.backpack: Inventory = Any
        self._unpatched_inventory: Callable[[BaseUser, Game], Coroutine[Any, Any, Inventory]]

    @register(EMsg.ClientFromGC)
    async def parse_gc_message(self, msg: MsgProto[CMsgGcClient]) -> None:
        if msg.body.appid != self.client._GAME.id:
            return

        try:
            language = self.__class__.Language(utils.clear_proto_bit(msg.body.msgtype))
        except ValueError:
            return log.info(
                f"Ignoring unknown msg type: {msg.body.msgtype} ({utils.clear_proto_bit(msg.body.msgtype)})"
            )

        try:
            gc_msg = (
                GCMsgProto(language, msg.body.payload)
                if utils.is_proto(msg.body.msgtype)
                else GCMsg(language, msg.body.payload)
            )
        except Exception as exc:
            return log.error("Failed to deserialize message: %r, %r", language, msg.body.payload, exc_info=exc)
        else:
            log.debug("Socket has received GC message %r from the websocket.", gc_msg)

        self.dispatch("gc_message_receive", gc_msg)
        self.run_parser(language, gc_msg)
